// Copyright (c) 2016, Herman Bergwerf. All rights reserved.
// Use of this source code is governed by an AGPL-3.0-style license
// that can be found in the LICENSE file.

part of omnitest.data;

/// These keys cannot be used as omnibusId for messages that are tested
/// with this program (seriously, why would you ever use these?).
const _reservedKeys = const ['expect', 'module-responses', 'mocked-responses'];

/// Get the first key in $map that can be identified as omnibusId given that
/// $ids contains all omnibusId's that can be used.
String _getOmnibusId(Iterable<String> allIds, YamlMap map) {
  var ids = map.keys.where((String key) => allIds.contains(key));
  return ids.length == 1 ? ids.first : null;
}

/// Safely cast a YamlNode to a Map.
Map<String, dynamic> _castYamlNodeToMap(YamlNode data) {
  return data is YamlMap
      ? pdsFromMap(data as Map<String, dynamic>)
      : new Map<String, dynamic>();
}

/// Data of a single message in a use case message chain
class OmnitestMessage {
  /// Message data
  OmnibusMessage data;

  /// Expected events
  List<OmnitestEvent> expectedEvents = new List<OmnitestEvent>();

  /// Expected requests
  List<OmnitestExpectedRequest> expectedRequests =
      new List<OmnitestExpectedRequest>();

  /// Expected (final) responses
  List<OmnitestResponse> expectedResponses = new List<OmnitestResponse>();

  /// Load data from a YAML map.
  OmnitestMessage.fromYaml(
      YamlMap yamlData, Map<String, OmnibuildMessage> messageSpec) {
    var omnibusId = _getOmnibusId(messageSpec.keys, yamlData);
    if (omnibusId != null) {
      // This cannot be a response itself (responses cannot be in the message
      // chain).
      if (messageSpec[omnibusId].type == OmnibusMessageType.response) {
        throw new OmnitestException.yaml(
            'A message-chain cannot contain responses.');
      }

      // Construct message data.
      if (messageSpec[omnibusId].type == OmnibusMessageType.event) {
        data = new OmnitestEvent(
            omnibusId, _castYamlNodeToMap(yamlData[omnibusId]));
      } else {
        data = new OmnitestRequest(
            omnibusId, _castYamlNodeToMap(yamlData[omnibusId]));
      }

      // Parse expected messages.
      if (yamlData.containsKey('expect') && yamlData['expect'] is YamlList) {
        yamlData['expect'].forEach((YamlMap yamlData) {
          String omnibusId = _getOmnibusId(messageSpec.keys, yamlData);
          if (omnibusId != null) {
            switch (messageSpec[omnibusId].type) {
              case OmnibusMessageType.event:
                // Parse message as event and store it in $expectedEvents.
                expectedEvents.add(new OmnitestEvent(
                    omnibusId, _castYamlNodeToMap(yamlData[omnibusId])));
                break;

              case OmnibusMessageType.response:
                // Parse message as response and store it in $expectedResponses.
                expectedResponses.add(new OmnitestResponse(
                    omnibusId, _castYamlNodeToMap(yamlData[omnibusId])));
                break;

              case OmnibusMessageType.request:
                // Requests also have expected responses, therefore they are
                // stored using the OmnitestRequest class.
                expectedRequests.add(new OmnitestExpectedRequest.fromYaml(
                    omnibusId, yamlData, messageSpec));
                break;
            }
          } else {
            throw new OmnitestException.noOmnibusIdKey(yamlData);
          }
        });
      }

      // Check for duplicate requests
      //
      // Duplicate events and responses are allowed. However, duplicate requests
      // would either point to stateful requests which are highly discouraged,
      // or a useless request repeat.
      Set<OmnibusRequest> reqs;
      for (var req in expectedRequests) {
        if (!reqs.contains(req.data)) {
          reqs.add(req.data);
        } else {
          // This request is already contained in $reqs; throw an error.
          throw new OmnitestException.yaml(
              'You cannot contain duplicate requests in expect.');
        }
      }
    } else {
      throw new OmnitestException.noOmnibusIdKey(yamlData);
    }
  }

  /// Find the first OmnitestExpectedRequest for the given OmnitestRequest.
  OmnitestExpectedRequest findRequest(OmnitestRequest request) {
    return expectedRequests.firstWhere(
        (OmnitestExpectedRequest r) => request == r.data,
        orElse: () => null);
  }
}

/// Expected request that is published as result of a message in a use case
/// messag chain.
class OmnitestExpectedRequest {
  /// Request data
  OmnitestRequest data;

  /// Request responses that should be provided by the module.
  List<OmnitestResponse> moduleResponses;

  /// Request responses that should be mocked by omnitest.
  List<OmnitestResponse> mockedResponses;

  /// Load data from a YAML map.
  OmnitestExpectedRequest.fromYaml(String omnibusId, YamlMap yamlData,
      Map<String, OmnibuildMessage> messageSpec) {
    // Construct message data.
    data =
        new OmnitestRequest(omnibusId, _castYamlNodeToMap(yamlData[omnibusId]));

    // Common function for processing module and mocked responses.
    void processResponses(YamlList list, bool mocked) {
      list.forEach((YamlMap yamlData) {
        String omnibusId = _getOmnibusId(messageSpec.keys, yamlData);
        if (omnibusId != null) {
          // Validate that this is a response message.
          if (messageSpec[omnibusId].type == OmnibusMessageType.response) {
            // Construct message.
            var message = new OmnitestResponse(
                omnibusId, _castYamlNodeToMap(yamlData[omnibusId]));

            // Add message to $moduleResponses or $mockedResponses.
            if (mocked) {
              mockedResponses.add(message);
            } else {
              moduleResponses.add(message);
            }
          } else {
            throw new OmnitestException.yaml(
                'Only response message are allowed in ${mocked ? 'mocked' : 'module'}-messages.');
          }
        } else {
          throw new OmnitestException.noOmnibusIdKey(yamlData);
        }
      });
    }

    // Loop through module responses.
    if (yamlData.containsKey('module-responses') &&
        yamlData['module-responses'] is YamlList) {
      processResponses(yamlData['module-responses'], false);
    }

    // Loop through mocked responses.
    if (yamlData.containsKey('mocked-responses') &&
        yamlData['mocked-responses'] is YamlList) {
      processResponses(yamlData['mocked-responses'], true);
    }
  }
}
