// Copyright (c) 2016, Herman Bergwerf. All rights reserved.
// Use of this source code is governed by an AGPL-3.0-style license
// that can be found in the LICENSE file.

part of omnitest.exceptions;

/// JSON encoder for writing fail messages.
final JsonEncoder _jsonEncoder = new JsonEncoder.withIndent('  ');

/// AnsiPen to write red text.
final redPen = new AnsiPen()..red();

/// Split [a] into lines and prefix them with [b]. Returns all lines joined
/// together using line endings.
String _prefixEachLine(String str, String prefix) {
  var out = new StringBuffer();
  var lines = LineSplitter.split(str);
  lines.forEach((String line) {
    out.writeln('$prefix$line');
  });
  return out.toString();
}

/// Error class used for internal error throwing.
class OmnitestException extends Error {
  /// Error message
  String message;

  /// Default constructor
  OmnitestException(this.message);

  /// Render YAML error
  OmnitestException.yaml(String error) {
    message = redPen('There was an error in your YAML test file:\n$error\n');
  }

  /// Render no omnibusId in message map found error
  OmnitestException.noOmnibusIdKey(YamlMap data) {
    message = redPen('''
There was an error in your YAML test file:
No valid omnibusId was found in the following message:
${_jsonEncoder.convert(data)}
''');
  }

  /// Render unexpected events error.
  OmnitestException.unexpectedEvents(
      List<OmnibusEvent> expected, List<OmnibusEvent> actual) {
    // Render message
    message = redPen(_msgUnexpectedEvents.renderString({
      'expectedEvents': _generateMustacheList(expected),
      'actualEvents': _generateMustacheList(actual)
    }));
  }

  /// Render unexpected responses error.
  OmnitestException.unexpectedResponses(OmnibusRequest request,
      List<OmnibusResponse> expected, List<OmnibusResponse> actual) {
    // Render message
    message = redPen(_msgUnexpectedResponses.renderString({
      'request': {
        'omnibusId': request.omnibusId,
        'jsonData':
            _prefixEachLine(_jsonEncoder.convert(request.toPds()), '    ')
      },
      'expectedResponses': _generateMustacheList(expected),
      'actualResponses': _generateMustacheList(actual)
    }));
  }

  /// Render unexpected request error.
  OmnitestException.unexpectedRequest(OmnibusRequest request) {
    message = redPen(_msgUnexpectedRequest.renderString({
      'omnibusId': request.omnibusId,
      'data': _prefixEachLine(_jsonEncoder.convert(request.toPds()), '    ')
    }));
  }

  /// toString should return the message.
  String toString() => message;

  /// Generate mustache list.
  List _generateMustacheList(List<OmnibusMessage> list) {
    var newList = new List<Map<String, String>>();
    list.forEach((OmnibusMessage item) {
      newList.add({
        'omnibusId': item.omnibusId,
        'jsonData': _prefixEachLine(_jsonEncoder.convert(item.toPds()), '    ')
      });
    });
    return newList;
  }
}

/// Template for _OmniError.unexpectedEvents
final _msgUnexpectedEvents = new Template('''
The module published unexpected events:
  Expected events:
  {{# expectedEvents }}
  - omnibusId: {{ omnibusId }}
    data: {{{ jsonData }}}
  {{/ expectedEvents }}
  Actual events:
  {{# actualEvents }}
  - omnibusId: {{ omnibusId }}
    data: {{{ jsonData }}}
  {{/ actualEvents }}
''');

/// Template for _OmniError.unexpectedResponses
final _msgUnexpectedResponses = new Template('''
The module published unexpected responses:
  Request:
    omnibusId: {{ request.omnibusId }}
    data:
{{{ request.jsonData }}}
  Expected responses from module:
  {{# expectedResponses }}
  - omnibusId: {{ omnibusId }}
    data:
{{{ jsonData }}}
  {{/ expectedResponses }}
  Actual responses from module:
  {{# actualResponses }}
  - omnibusId: {{ omnibusId }}
    data:
{{{ jsonData }}}
  {{/ actualResponses }}
''');

/// Template for _OmniError.unexpectedRequest
final _msgUnexpectedRequest = new Template('''
A request was published but not expected:
  omnibusId: {{ omnibusId }}
  data: {{{ jsonData }}}
''');
