// Copyright (c) 2016, Herman Bergwerf. All rights reserved.
// Use of this source code is governed by an AGPL-3.0-style license
// that can be found in the LICENSE file.

part of omnitest;

/// Module bootstrap function
typedef void ModuleMain(Omnibus bus, http.BaseClient httpClient);

/// Function that is passed down to the test framework.
typedef Future OmnitestFunction();

/// Create test function
///
/// The returned function can be passed to the test function from the test
/// package. The function will execute all tests that are described in the
/// [testFile] against the [module].
OmnitestFunction omnitest(ModuleMain module, String testFile) {
  Future run() async {
    // Parse YAML config.
    OmnitestFile tests;
    try {
      tests = new OmnitestFile();
      await tests.parseYaml(new File(testFile).readAsStringSync());
    } on OmnitestException catch (e) {
      fail(e.message);
    }

    // Loop through use cases.
    for (var i = 0; i < tests.useCases.length; i++) {
      await runZoned(() async {
        // Used to check if a given request is expected by the pending message.
        OmnitestMessage currentMessage;

        // Important note on stateful requests
        // -----------------------------------
        // The usage of stateful requests cannot be tested with omnitest since
        // you can only define a specific request once per message in the
        // message chain, and not multiple times with different responses. This
        // is by design, your code should never have to rely on state data in
        // the modules behind the messages.

        // Create new test fabric.
        var fabric = new _OmnitestFabric(
            (OmnitestRequest request, List<OmnitestResponse> moduleResponses) {
          OmnitestExpectedRequest requestInformation =
              currentMessage.findRequest(request);
          if (requestInformation != null) {
            // Check if the module responses are as expected.
            if (!_orderlessListCompare(
                moduleResponses, requestInformation.moduleResponses)) {
              throw new OmnitestException.unexpectedResponses(
                  request, requestInformation.moduleResponses, moduleResponses);
            }

            // Return the mocked responses.
            return requestInformation.mockedResponses;
          } else {
            throw new OmnitestException.unexpectedRequest(request);
            return [];
          }
        });

        // Bootstrap module.
        module(fabric.bus, tests.useCases[i].httpClient);

        // Run message chain.
        for (var message in tests.useCases[i].messageChain) {
          currentMessage = message;
          await _testMessage(message, fabric);
        }
      }, onError: (e, stackTrace) {
        var nr = i + 1;
        fail('''
In testcase #$nr:
Description: ${tests.useCases[i].description}

$e

$stackTrace
''');
      });
    }
  }
  return run;
}

/// Evaluate the outcome of a single message on the given fabric.
Future _testMessage(OmnitestMessage message, _OmnitestFabric fabric) async {
  // Clear event log so only new events are collected.
  fabric.eventLogger.log.clear();

  // Publish message.
  if (message.data.type == OmnibusMessageType.event) {
    // Await all event handlers.
    await (await fabric.bus.publishEvent(message.data)).onComplete();
  } else {
    // Block the processing of the tested request by the fabric.
    fabric.requestHandler.skipOnce();

    // Await all responses.
    var responses =
        await (await fabric.bus.publishRequest(message.data)).onComplete();

    // Convert all responses to OmnitestResponse to make them comparable.
    var omnitestResponses = new List<OmnitestResponse>();
    responses.forEach((OmnibusResponse response) => omnitestResponses
        .add(new OmnitestResponse(response.omnibusId, response.toPds())));

    // Check responses.
    if (!_orderlessListCompare(message.expectedResponses, omnitestResponses)) {
      throw new OmnitestException.unexpectedResponses(
          message.data, message.expectedResponses, responses);
    }
  }

  // Check event log.
  // Note that requests are already instantly validated inside the global
  // request handler fabric.
  if (!_orderlessListCompare(message.expectedEvents, fabric.eventLogger.log)) {
    throw new OmnitestException.unexpectedEvents(
        message.expectedEvents, fabric.eventLogger.log);
  }
}
